/* -*-  mode: c++; c-basic-offset: 4; -*- */
#ifndef EVAL_STACK_HH
#define EVAL_STACK_HH

#include <iosfwd>
#include <string>
#include <vector>

class eval_token {
public:
    typedef unsigned int datype;

public:
    virtual ~eval_token(void) {}
    virtual eval_token* clone(void) const = 0;
    virtual datype getType() const = 0;
    virtual bool getLogical() const;
    virtual double getNumeric() const;
    virtual std::string getString() const = 0;
    virtual void setLogical(bool yorn);
    virtual void setNumeric(double val);
    virtual void setString(const std::string& s) = 0;
};

/**  The evaluation stack holds currently active intermediate results.
  *  The stack items are store as pointers to eval_token objects.
  *  @memo Evaluation stack.
  *  @author John Zweizig
  *  @version 1.3; Last modified March 14, 2006
  */
class eval_stack {

public:
    /** Operation code enumeration.
      */
    enum OpsEnum {
	kNoop,
	kPushNumer,
	kPushConst,
	kPushSymbol,
	kCallFc,
	kAdd,
	kSub,
	kMpy,
	kDiv,
	kOr,
	kAnd,
	kLogAnd,
	kLogOr,
	kCmpEq,
	kCmpNE,
	kCmpLt,
	kCmpLE,
	kCmpGt,
	kCmpGE,
	kModulo,
	//-----------   kNOps (number of operators) must be last entry.
	kNOps
    };

    typedef int op_level;

public:
    /**  Evaluation stack constructor.
     */
    eval_stack(void);

    /** Evaluation stacke destructor. 
      */
    virtual ~eval_stack(void);

    /**  Clear the evaluation stack.
     */
    void clear(void);

    /**  Print out the contents of the evaluation stack.
     */
    virtual void dump(std::ostream& out) const;

    /**  Test whether there are any data on the stack.
     */
    bool empty(void) const;

    /** perform an operation on the top stack entries.
     */
    virtual void evaluate(OpsEnum op);

    /**   Get the operator ID>
      */
    virtual OpsEnum getOpID(const std::string& op) const;

    /**  Get the operator level (priority).
      */
    virtual op_level getOpLevel(OpsEnum op) const;

    /**  Get the operator symbol.
      */
    virtual const char* getOpSym(OpsEnum op) const;

    /**  Look at a datum on the stack. The datum position is specified 
     *  relative to the top of the stack, with #inx=1# indicating the 
     *  top entry, #inx=2# indicates the next to the top, etc.
     */
    const eval_token& peek(int i=1) const;

    /** Push a token on the top of the stack.
     */
    virtual void push(const eval_token& tkn);

    /** Push a token on the top of the stack.
     */
    virtual void push(eval_token* tkn);

    /** Push a string on the top of the stack.
     */
    virtual void push(const std::string& str);

    /**  Push a numeric value. The double precision float value is converted
     *  to a string before being pushed on to the stack.
     */
    virtual void push(double val);

    /**  Push a boolean value onto the stack.
     */
    virtual void push(bool val);

    /** Remove the top token off the stack and delete it.
     */
    void remove(void);

    /** Get the stack size.
     */
    unsigned int size(void) const;

    /**  Pop the top value off the stack and convert it to a boolean value.
     *  Non-zero numeric values and the string #"true"# are considered to
     *  be true, and all other values are considered to be false.
     */
    virtual bool pop_logical(void);

    /**  Pop the top value off the stack and convert it to a numeric value.
     */
    virtual double pop_numeric(void);

    /**  Pop the top value off the stack and return the string value.
     */
    virtual std::string pop_string(void);

    /**  Pop the top value off the stack and return the string value.
     */
    virtual eval_token* pop_token(void);

    /**  Get a reference to the top element of the stack.
     */
    virtual eval_token& top(void);
    virtual const eval_token& top(void) const;

    /**  Get the data type of the specified stack entry. Note that the top
     *  stack entry is off=1.
     */
    eval_token::datype type(int off=1) const;

private:
    op_level mOpLevel[kNOps];
    const char* mOpStr[kNOps];

private:
    typedef std::vector< eval_token* > eval_vect;
    eval_vect mVect;
};

//======================================  Inline methods
inline bool 
eval_stack::empty(void) const {
    return mVect.empty();
}

inline eval_token*
eval_stack::pop_token(void) {
    eval_token* p = mVect.back();
    mVect.pop_back();
    return p;
}

inline void
eval_stack::remove(void) {
    delete pop_token();
}

inline void
eval_stack::clear(void) {
    while (!empty()) remove();
}

inline unsigned int
eval_stack::size(void) const {
    return mVect.size();
}

inline const eval_token&
eval_stack::top(void) const {
    return *mVect.back();
}

inline eval_token&
eval_stack::top(void) {
    return *mVect.back();
}

#endif // !defined(EVAL_STACK_HH)
