/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "DMTGen.hh"
#include "ParseLine.hh"
#include "FilterDesign.hh"
#include "MultiPipe.hh"
#include "Delay.hh"
#include "DaqErrFilter.hh"
#include "IfoResponse.hh"
#include "FDPipe.hh"
#include "FDResample.hh"
#include <unistd.h>
#include <iostream>

using namespace std;
using namespace generator;

//======================================  Main function
int
main(int argc, const char* argv[]) {
    DMTGen myPgm(argc, argv);
    if (!myPgm.configured()) return 1;
    myPgm.generate();
    return 0;
}

//======================================  Main function
DMTGen::DMTGen(int argc, const char* argv[]) 
  : mConfigured(false), mWriter("DMTGen", mRunID), mDebug(0), mNFramesOut(0)
{
    bool syntax(false);
    string config;
    for (int i=1 ; i < argc ; ++i) {
        string argi = argv[i];
        if (argi == "-conf" ) {
	    config = argv[++i];
	} else if (argi == "-debug" ) {
	    if (i+1<argc && *argv[i+1] != '-') mDebug = strtol(argv[++i], 0,0);
	    else                               mDebug = 1;
	} else {
	    syntax = true;
	}
    }

    //----------------------------------  Set up parameter list
    mDict.addPar("Identifier",  "DMT generation run");
    mDict.addPar("RunID",       -666);
    mDict.addPar("FilePrefix",  "DMTGen");
    mDict.addPar("StartGPS",       0);
    mDict.addPar("EndGPS",         0);
    mDict.addPar("GenerStride",  4.0);
    mDict.addPar("FrameLength", 16.0);
    mDict.addPar("Compression",   "");
    mGener.setDebug(mDebug);

    //----------------------------------  Get the configuration
    if (config.empty()) {
        cerr << "Configuration file not specified." << endl;
	syntax = true;
    }
    if (syntax) {
        cerr << "Command line syntax:" << endl;
	cerr << "  DMTGen -conf <config-file> [-debug <dbg-level>]" << endl;
	return;
    }
    if (readConfig(config)) {
        cerr << "Error reading configuration file: " << config << endl;
	return;
    }
    mGener.setGenerStride(mDict.getDouble("GenerStride"));

    //----------------------------------  Check that the start time is valid
    mStartTime = Time(mDict.getInt("StartGPS"));
    if (!mStartTime) {
        cerr << "Start time not specified" << endl;
	return;
    }

    mDt = mDict.getDouble("FrameLength");
    if (mDt <= Interval(0.0)) {
        cerr << "Frame length not positive" << endl;
	return;
    }

    mEndTime = Time(mDict.getInt("EndGPS"));
    if (!mEndTime) {
        mEndTime = mStartTime + mDt;
    } else if (mEndTime < mStartTime) {
        cerr << "Run ends before start" << endl;
	return;
    }

    mRunID   = mDict.getInt("RunID");
    mIdent   = mDict.getString("Identifier");
    mFilePfx = mDict.getString("FilePrefix");
    mWriter.setRunID(mRunID);
    mWriter.setName(mIdent);
    string compression = mDict.getString("Compression");
    if (compression == "gzip") {
        mWriter.setCompress(FrVectRef::kGZip);
    } else  if (compression == "zero-suppress-2") {
        mWriter.setCompress(FrVectRef::kZeroSuppress2);
    } else  if (compression == "zero-suppress-4") {
        mWriter.setCompress(FrVectRef::kZeroSuppress4);
    } else if (!compression.empty()) {
        cout << "Unrecognized compression mode: " << compression 
	     << " Ignored!" << endl;
    }

    //----------------------------------  All done;
    mConfigured = true;
}

//======================================  Destructor
DMTGen::~DMTGen(void) {
    mGener.print_stats(cout);
    cout << "Number of frames written = " << mNFramesOut << endl;
    cout << "DMTGen Completed." << endl;
}

//======================================  Read a configuration file.
bool
DMTGen::readConfig(const std::string& file) {
    ParseLine pl(file.c_str());
    if (!pl.isOpen()) return true;
    pl.setLog(cout);

    //----------------------------------  Loop over commands.
    bool syntax = false;
    while (pl.getLine() >= 0) {
        int nArgs = pl.getCount();
	if (!nArgs) continue;
	string cmd = pl[0];

	//------------------------------  Specify a channel to be generated.
	if (cmd == "Channel") {
	    int cid = mGener.addChannel(pl[1]);

	    //--------------------------  Loop over global modifiers
	    int i=2;
	    for ( ; i<nArgs && *(pl[i]) == '-' ; ++i) {
	        string argi = pl[i];
		if (argi == "-daqchain") {
		    const_filt_iter pF=mFilterMap.find(pl[++i]);
		    if (pF == mFilterMap.end()) {
		        cout << "Unrecognized filter name" << endl;
			syntax = true;
		    } else {
		        mGener.refChannel(cid).setDaqChain(pF->second);
		    }
		} else if (argi == "-sample") {
		    mGener.refChannel(cid).setSample(pl.getDouble(++i));
		} else {
		    cout << "Unrecognized argument: " << argi << endl;
		    syntax = true;
		}
	    }
	    
	    //--------------------------  Loop over source specifiers
	    for ( ; i<nArgs ; ++i) {
	        const char* p  = pl[i];
		const char* p0 = p;
		string name;
		const Pipe* filter=0;
		MultiPipe*  mpfilt=0;
		for ( ; *p ; p++) {
		    if (*p == '|') {
		        if (p0 == pl[i]) {
			    name = string(p0, p-p0);
			    p0 = p + 1;
			} else {
			    const_filt_iter pF=mFilterMap.find(string(p0, p-p0));
			    if (pF == mFilterMap.end()) {
			        cout << "Unidentified filter: " 
				     << string(p0, p-p0)
				     << endl;
				syntax = true;
				break;
			    }
			    if (!filter) {
			        filter = pF->second;
			    } else {
			        if (!mpfilt) {
				    mpfilt = new MultiPipe;
				    mpfilt->addPipe(*filter);
				}
				mpfilt->addPipe(*(pF->second));
			    }
			}
		    }
		}

		//----------------------  handle the string after the last |
		if (p0 == pl[i]) {
		    name = string(p0, p-p0);
		} else {
		    const_filt_iter pF=mFilterMap.find(string(p0, p-p0));
		    if (pF == mFilterMap.end()) {
		        cout << "Unidentified source: " 
			     << string(p0, p-p0)
			     << endl;
			syntax = true;
			continue;
		    }
		    if (!filter) {
		        filter = pF->second;
		    } else {
		        if (!mpfilt) {
			    mpfilt = new MultiPipe;
			    mpfilt->addPipe(*filter);
			}
			mpfilt->addPipe(*(pF->second));
		    }
		}

		//----------------------  Add the source descriptor
		int sid = mGener.getSourceID(name);
		if (sid < 0) {
		    cerr << "Unidentified source: " << name << endl;
		    syntax = true;
		} else if (mpfilt) {
		    mGener.refChannel(cid).addComponent(sid, mpfilt);
		} else {
		    mGener.refChannel(cid).addComponent(sid, filter);
		}
		delete mpfilt;
	    }

	//------------------------------  Add a filter to the database.
	} else if (cmd == "Filter") {
	    if (nArgs < 3) {
	        cout << "Syntax error in filter definition: <3 arguments"
		     << endl;
		cout << "    Filter description syntax is: "
		     << "'Filter <name> <type> <arguments>'" << endl;
		syntax = true;
		continue;
	    }

	    //--------------------------  Don't allow redefinition.
	    string name=pl[1];
	    if (mFilterMap.find(name) != mFilterMap.end()) {
	        cout << "Filter: " << name << " is already defined." <<endl;
		syntax = true;
		continue;
	    }

	    //--------------------------  Create Daq Error 'filter'.
	    string typ=pl[2];
	    Pipe*  fPipe(0);
	    if (typ == "DaqErr") {
	        if (nArgs < 4) {
		    cout << "Error in Filter specification. Syntax is:" << endl
			 << "Filter <name> DaqErr <type> [-clust <width>]" 
			 << "[-avgwid <width>] [-rate <err-rate>]" << endl;
		    syntax = true;
		    continue;
		}
		DaqErrFilter temp(pl[3]);
		for (int j=4; j<nArgs; j++) {
		    string argj = pl[j];
		    if (argj == "-clust") {
		        temp.setClusterWidth(pl.getInt(++j));
		    } else if (argj == "-avgwid") {
		        temp.setAvgWidth(pl.getDouble(++j));
		    } else if (argj == "-rate") {
		        temp.setErrRate(pl.getDouble(++j));
		    } else {
		        cout << "Unrecognized argument: " << argj << endl;
			syntax = true;
			break;
		    }
		}
		fPipe = temp.clone();

	    //--------------------------  Create a time delay 'filter'
	    } else if (typ == "Delay") {
	        if (nArgs < 4) {
		    cout << "Error in Filter specification. Syntax is: "
			 << "Filter <name> Delay <delay-time>" << endl;
		    syntax = true;
		    continue;
		}
		fPipe = new Delay(pl.getDouble(3));

	    //--------------------------  Design a filter
	    } else if (typ == "Design") {
	        if (nArgs < 5 || pl.getDouble(3) <= 0.0) {
		    cout << "Error in Filter specification. Syntax is: "
			 << "Filter <name> Design <sample-rate> <descriptor>"
			 << endl;
		    syntax = true;
		    continue;
		}
		FilterDesign fd(pl[4], pl.getDouble(3));
		fPipe = fd.release();

	    //--------------------------  f-Domain resampling filter
	    } else if (typ == "Resample") {
	        if (nArgs < 4 || pl.getDouble(3) <= 0.0) {
		    cout << "Error in Filter specification. Syntax is: "
			 << "Filter <name> Resample <new-sample-rate>"
			 << endl;
		    syntax = true;
		    continue;
		}
		FDResample fdr(pl.getDouble(3));
		fPipe = new FDPipe(fdr, FDPipe::overlap);

	    //--------------------------  f-Domain calibrated response filter
	    } else if (typ == "IfoResponse") {
	        if (nArgs < 4) {
		    cout << "Error in Filter specification. Syntax is: "
			 << "Filter <name> IfoResponse <xml-calib>"
			 << endl;
		    syntax = true;
		    continue;
		} else if (access(pl[3], R_OK)) {
		    cout << "Unable to access file: " << pl[3] << endl;
		    syntax = true;
		    continue;
		}
		IfoResponse fdr("Calibration", pl[3]);
		fPipe = new FDPipe(fdr, FDPipe::overlap);

	    //--------------------------  Unknown filter type
	    } else {
	        cout << "Unrecognized filter type: " << typ << endl;
		syntax = true;
		continue;
	    }

	    //--------------------------  Add the design to the database.
	    if (fPipe) mFilterMap[name] = fPipe;

	//------------------------------  Include a configuration file
	} else if (cmd == "Include") {
	    if (nArgs < 2) {
	        cerr << "Missing file name. Include syntax: " << endl;
	        cerr << "   Include <file-path>" << endl;
	        syntax = true;
	    } else {
	        syntax = readConfig(pl[1]);
	    }

	//------------------------------  Set specified parameters
	} else if (cmd == "Parameter") {
	    if (mDict.exists(pl[1])) {
	        mDict.setPar(pl[1], pl[2]);
	    } else {
	        cerr << "Parameter " << pl[1] << " isn't defined." << endl;
		syntax = true;
	    }


	} else if (cmd == "Source") {
	    DataSource* src = DataSource::parse(pl[2]);
	    if (!src) {
	        cerr << "Unparseable source description - source " 
		     << pl[1] << " not defined." << endl;
		syntax = true;
		continue;
	    }
	    int sid = mGener.addSource(pl[1], src);
	    for (int i=3 ; i<nArgs ; ++i) {
	        string argi = pl[i];
		if (argi == "-rate") {
		    double rate = pl.getDouble(++i);
		    mGener.refSource(sid).setTriggerRate(rate);
		} else if (argi == "-sample") {
		    double rate = pl.getDouble(++i);
		    mGener.refSource(sid).setSample(Interval(1./rate));
		} else if (argi == "-simdata") {
		    mGener.refSource(sid).setWriteFlag(true);
		} else if (argi == "-simevent") {
		    mGener.refSource(sid).setSaveFlag(true);
		} else if (argi == "-time") {
		    double t0 = pl.getDouble(++i);
		    long gps = long(t0);
		    mGener.refSource(sid).setTriggerTime(Time(gps)+Interval(t0-gps));
		} else {
		    cerr << "Unidentified argument: " << argi << endl;
		    syntax = true;
		    break;
		}
	    }
	} else {
	    cerr << "Undefined command: " << cmd << endl;
	    syntax = true;
	}
    }
    return syntax;
}

//======================================  Generate the data.
void
DMTGen::generate(void) {

    //----------------------------------  Loop over gneration strides.
    Time writeTime = mStartTime;
    Interval dtGen = mDict.getDouble("GenerStride");
    for (Time now=mStartTime ; now < mEndTime ; now += dtGen) {
        if (mDebug) cout << "Generating data for GPS " << now.getS()
			 << " (generation length=" << dtGen << ")" << endl;
        mGener.generate(now, dtGen);
        if (mDebug) cout << "Data generated to " << mGener.getLatest().getS()
			 << endl;
	if (mGener.getLatest() >= writeTime + mDt) {
	    if (mDebug) cout << "Writing data for GPS " << now.getS()
			     << " (frame length=" << mDt << ")" << endl;
	    writeFrame(writeTime, mDt);
	    writeTime = writeTime += mDt;
	    mGener.releaseSourceData(writeTime);
	    mGener.releaseChannelData(writeTime);
	}
    }

    //----------------------------------  Write out (short?) last frame
    if (writeTime < mEndTime) {
        mGener.generate(mEndTime, Interval(0));
	Interval dt = mGener.getLatest() - writeTime;
	if (double(dt) > 0) writeFrame(writeTime, dt);
	writeTime += dt;
	mGener.releaseSourceData(writeTime);
	mGener.releaseChannelData(writeTime);
    }
}

//======================================  Write generated data
void 
DMTGen::writeFrame(const Time& t0, Interval dT) {

    //----------------------------------  Start the frame
    mWriter.buildFrame(t0, dT);
    mWriter.addHistory("DMTGener", Now(), mIdent);

    //----------------------------------  Write out selected simulated data
    int  nSrc  = mGener.getNSource();
    for (int i=0 ; i < nSrc ; ++i) {
        DataSource& src(mGener.refSource(i));
	if (src.getSaveFlag()) {
	    unsigned long N=src.getNEvent();
	    unsigned long nCopied=0;
	    for (unsigned long j=0 ; j<N ; j++) {
	        const DataSource::SrcEvent& sev(src.refEvent(j));
	        if (sev.getGPSMax() >= t0 + dT) break;
		mWriter.addSimEvent(sev.getName(), 
				    sev.getComment(), 
				    sev.getInputs(), 
				    sev.getGPSMax(),
				    sev.getTimeBefore(), 
				    sev.getTimeAfter(), 
				    sev.getAmplitude(),
				    sev.getParamList()
				    );
		nCopied++;
	    }
	    src.eraseEvent(nCopied);
	}
	if (src.getWriteFlag()) {
	    mWriter.addSimSeries(src.getName(), src.getTimeSeries(t0,dT));
	}
    }

    //----------------------------------  Write out data channels
    int  nChan = mGener.getNChannel();
    string dList;
    for (int i=0 ; i < nChan ; ++i) {
        GenChan& chan(mGener.refChannel(i));
	string cName = chan.getChannel();
	if (dList.find(cName.substr(0,3)) == string::npos) {
	    dList += cName.substr(0,3);
	    mWriter.addStdDetector(cName.c_str());
	    if (mDebug) cout << "Adding detector: "<< cName.substr(0,2)<< endl;
	}
	mWriter.addRawSeries(chan.getChannel(),chan.response().extract(t0,dT));
	if (mDebug) cout << "  writing channel " << chan.getChannel()
			 << " (" << chan.response().getNSample() 
			 << " samples)" << endl; 
    }

    //----------------------------------  Construct the output file name
    ostringstream ss;
    long begTgps = t0.getS();
    Time endT = t0+dT;
    long endTgps = endT.getS();
    if (endT.getN()) ++endTgps;
    ss << mFilePfx << "-" << begTgps << "-" << endTgps - begTgps << ".gwf";

    //----------------------------------  Write a frame;
    if (mDebug) cout << "  opening file: " << ss.str() << endl;
    if (mWriter.open(ss.str().c_str(), true)) {
        mWriter.erase();
	cerr << "DMTGen: Failed to open frame " << ss.str() << " for output." 
	     << endl;

    } else {
        mWriter.writeFrame();
	mWriter.close();
	mNFramesOut++;
    }
}
